from compas.geometry import Point, distance_point_point_sqrd
from compas.utilities.itertools import pairwise
from compas_slicer.slicers.slice_utilities import create_graph_from_mesh_edges, sort_graph_connected_components
import compas_slicer.utilities as utils
import logging
from abc import abstractmethod
from compas_slicer.geometry import Path

logger = logging.getLogger('logger')

__all__ = ['ContoursBase']


class ContoursBase(object):
    """
    This is meant to be extended by all classes that generate isocontours of a scalar function on a mesh.
    This class handles the two steps of iso-contouring of a triangular mesh consists of two steps;
    1)find intersected edges and 2)sort intersections using a graph to generate coherent polylines.

    The inheriting classes only have to implement the test that checks if an edge is intersected,
    and the method to find the zero crossing of an intersection.

    Attributes
    ----------
    mesh : :class: 'compas.datastructures.Mesh'

    """

    def __init__(self, mesh):
        self.mesh = mesh
        self.intersection_data = {}  # dict: (ui,vi) : {compas.Point}
        # key: tuple (int, int), The edge from which the intersection point originates.
        # value: :class: 'compas.geometry.Point', The zero-crossing point.
        self.edge_to_index = {}  # dict that stores node_index and edge relationship
        # key: tuple (int, int) edge
        # value: int, index of the intersection point
        self.sorted_point_clusters = {}  # dict
        # key: int, The index of the connected component
        # value: list, :class: 'compas.geometry.Point', The sorted zero-crossing points.
        self.sorted_edge_clusters = {}  # dict
        # key: int, The index of the connected component.
        # value: list, tuple (int, int), The sorted intersected edges.
        self.closed_paths_booleans = {}  # dict
        # key: int, The index of the connected component.
        # value: bool, True if path is closed, False otherwise.

    def compute(self):
        self.find_intersections()
        G = create_graph_from_mesh_edges(self.mesh, self.intersection_data, self.edge_to_index)
        sorted_indices_dict = sort_graph_connected_components(G)

        nodeDict = dict(G.nodes(data=True))
        for key in sorted_indices_dict:
            sorted_indices = sorted_indices_dict[key]
            self.sorted_edge_clusters[key] = [nodeDict[node_index]['mesh_edge'] for node_index in sorted_indices]
            self.sorted_point_clusters[key] = [self.intersection_data[e] for e in self.sorted_edge_clusters[key]]

        self.label_closed_paths()

    def label_closed_paths(self):
        for key in self.sorted_edge_clusters:
            first_edge = self.sorted_edge_clusters[key][0]
            last_edge = self.sorted_edge_clusters[key][-1]
            u, v = first_edge
            self.closed_paths_booleans[key] = u in last_edge or v in last_edge

    def find_intersections(self):
        """
        Fills in the
        dict self.intersection_data: key=(ui,vi) : [xi,yi,zi],
        dict self.edge_to_index: key=(u1,v1) : point_index. """
        for edge in list(self.mesh.edges()):
            if self.edge_is_intersected(edge[0], edge[1]):
                point = self.find_zero_crossing_data(edge[0], edge[1])
                if point:  # Sometimes the result can be None
                    if edge not in self.intersection_data and tuple(reversed(edge)) not in self.intersection_data:
                        # create [edge - point] dictionary
                        self.intersection_data[edge] = {}
                        self.intersection_data[edge] = Point(point[0], point[1], point[2])

            # create [edge - point] dictionary
            for i, e in enumerate(self.intersection_data):
                self.edge_to_index[e] = i

    def save_point_clusters_as_polylines_to_json(self, DATA_PATH, name):
        all_points = {}
        for i, key in enumerate(self.sorted_point_clusters):
            all_points[i] = utils.point_list_to_dict(self.sorted_point_clusters[key])
        utils.save_to_json(all_points, DATA_PATH, name)

    # --- Abstract methods
    @abstractmethod
    def edge_is_intersected(self, u, v):
        """ Returns True if the edge u,v has a zero-crossing, False otherwise. """
        # to be implemented by the inheriting classes
        pass

    @abstractmethod
    def find_zero_crossing_data(self, u, v):
        """ Finds the position of the zero-crossing on the edge u,v. """
        # to be implemented by the inheriting classes
        pass

    def add_to_vertical_layers_manager(self, vertical_layers_manager):
        for key in self.sorted_point_clusters:
            pts = self.sorted_point_clusters[key]
            if len(pts) > 3:  # discard curves that are too small
                path = Path(pts, is_closed=self.closed_paths_booleans[key])

                vertical_layers_manager.add(path)

    @property
    def is_valid(self):
        if len(self.sorted_point_clusters) > 0:
            for key in self.sorted_point_clusters:
                pts = self.sorted_point_clusters[key]
                if len(pts) > 3:
                    path_total_length = 0.0
                    for p1, p2 in pairwise(pts):
                        path_total_length += distance_point_point_sqrd(p1, p2)
                    if path_total_length > 1.0:
                        return True  # make sure there is at least one path with acceptable length
        return False



from compas.geometry import norm_vector, Vector
import logging

logger = logging.getLogger('logger')

__all__ = ['set_blend_radius']


def set_blend_radius(print_organizer, d_fillet=10, buffer=0.3):
    """Sets the blend radius (filleting) for the robotic motion.

    Parameters
    ----------
    print_organizer: :class:`compas_slicer.slicers.BasePrintOrganizer`
    d_fillet: float
        Value to attempt to fillet with. Defaults to 10 mm.
    buffer: float
        Buffer to make sure that the blend radius is never too big.
        Defaults to 0.3.
    """

    logger.info("Setting blend radius")

    for printpoint, i, j, k in print_organizer.printpoints_indices_iterator():
        layer_key = 'layer_%d' % i
        path_key = 'path_%d' % j
        neighboring_items = print_organizer.get_printpoint_neighboring_items(layer_key, path_key, k)

        radius = d_fillet
        if neighboring_items[0]:
            radius = min(radius, norm_vector(Vector.from_start_end(neighboring_items[0].pt, printpoint.pt)) * buffer)

        if neighboring_items[1]:
            radius = min(radius, norm_vector(Vector.from_start_end(neighboring_items[1].pt, printpoint.pt)) * buffer)

        radius = round(radius, 5)

        printpoint.blend_radius = radius


if __name__ == "__main__":
    pass



import logging
import math
from compas_slicer.parameters import get_param
from compas.geometry import Point, Vector
from compas_slicer.geometry import PrintPoint
from datetime import datetime

logger = logging.getLogger('logger')

__all__ = ['create_gcode_text']


def create_gcode_text(print_organizer, parameters):
    """ Creates a gcode text file
    Parameters
    ----------
    print_organizer: :class: compas_slicer.print_organization.PrintOrganizer
    parameters : dict with gcode parameters.
        The defaults for those parameters are in the file compas_slicer.parameters.defaults_gcode.
    Returns
    ----------
    str, gcode text file
    """
    n_l = chr(10)  # new line
    # get time stamp
    now = datetime.now()
    datetimestamp = now.strftime("%H:%M:%S - %d %B %Y")
    logger.info('Generating gcode')
    gcode = ''

    #######################################################################
    # get all the necessary parameters:
    # Physical parameters
    # nozzle_diameter = get_param(parameters, key='nozzle_diameter', defaults_type='gcode')  # in mm
    filament_diameter = get_param(parameters, key='filament diameter', defaults_type='gcode')  # in mm

    # Dimensional parameters
    path_width = get_param(parameters, key='layer_width', defaults_type='gcode')  # in mm

    # Temperature parameters
    extruder_temperature = get_param(parameters, key='extruder_temperature', defaults_type='gcode')  # in °C
    bed_temperature = get_param(parameters, key='bed_temperature', defaults_type='gcode')  # in °C
    fan_speed = get_param(parameters, key='bed_temperature', defaults_type='gcode')  # 0-255
    fan_start_z = get_param(parameters, key='fan_start_z', defaults_type='gcode')  # in mm

    # Movement parameters
    feedrate = get_param(parameters, key='feedrate', defaults_type='gcode')  # in mm/s
    feedrate_travel = get_param(parameters, key='feedrate_travel', defaults_type='gcode')  # in mm/s
    feedrate_low = get_param(parameters, key='feedrate_low', defaults_type='gcode')  # in mm/s, for z < min_over_z
    feedrate_retraction = get_param(parameters, key='feedrate_retraction', defaults_type='gcode')  # in mm/s
    acceleration = get_param(parameters, key='acceleration', defaults_type='gcode')  # in mm/s²; ignored if 0
    jerk = get_param(parameters, key='jerk', defaults_type='gcode')  # in mm/s; if 0, the default driver value is used

    # Retraction and hop parameters
    z_hop = get_param(parameters, key='z_hop', defaults_type='gcode')  # in mm
    retraction_length = get_param(parameters, key='retraction_length', defaults_type='gcode')  # in mm
    retraction_min_travel = get_param(parameters, key='retraction_min_travel', defaults_type='gcode')  # in mm

    # Adhesion parameters
    flow_over = get_param(parameters, key='flow_over', defaults_type='gcode')  # as fraction > 1
    min_over_z = get_param(parameters, key='min_over_z', defaults_type='gcode')  # in mm
    # ______________________________________________________________________/ get parmeters

    # ######################################################################
    # gcode header
    gcode += ";Gcode with compas_slicer " + n_l
    gcode += ";Ioana Mitropolou <mitropoulou@arch.ethz.ch> @ioanna21" + n_l
    gcode += ";Joris Burger     <burger@arch.ethz.ch>      @joburger" + n_l
    gcode += ";Andrei Jipa      <jipa@arch.ethz.ch         @stratocaster>" + n_l
    gcode += ";MIT License" + n_l
    gcode += ";" + n_l
    gcode += ";generated " + datetimestamp + n_l
    gcode += ";" + n_l
    gcode += "T0                              ;set tool" + n_l  # for printing with multiple nozzles this will be useful
    gcode += "G21                             ;metric values" + n_l
    gcode += "G90                             ;absolute positioning" + n_l
    gcode += "M107                            ;start with the fan off" + n_l
    gcode += "M140 S" + str(bed_temperature) + "                        ;set bed temperature fast" + n_l
    gcode += "M104 S" + str(extruder_temperature) + "                       ;set extruder temperature fast" + n_l
    gcode += "M109 S" + str(extruder_temperature) + "                       ;set extruder temperature and wait" + n_l
    gcode += "M190 S" + str(bed_temperature) + "                        ;set bed temperature and wait" + n_l
    gcode += "G21                             ;metric values" + n_l
    gcode += "G90                             ;absolute positioning" + n_l
    gcode += "M83                             ;set e-values to relative while in absolute mode" + n_l
    if acceleration != 0:
        gcode += "M201 X" + str(acceleration) + " Y" + str(acceleration) + "          ;set max acceleration in xy" + n_l
    if jerk != 0:
        gcode += "M207 X" + str(jerk) + "            ;set max jerk" + n_l  # TODO: check firmware compatibility of M207
    gcode += "G28 X0 Y0                       ;home x and y axes" + n_l
    gcode += "G28 Z0                          ;home z axis independently" + n_l
    gcode += "G1 F4500                        ;set feedrate to 4,500 mm/min (75 mm/s)" + n_l
    gcode += "G1 Z15.0                        ;move nozzle up 15mm" + n_l
    gcode += "G1 F140 E29                     ;extruded slowly some filament (default: 29mm)" + n_l
    gcode += "G92 E0                          ;reset the extruded length" + n_l  # useless after M83, otherwise needed
    gcode += "G1 F" + str(feedrate_travel) + "                       ;set initial Feedrate" + n_l
    gcode += "M117 compas gcode print...      ;show up text on LCD" + n_l
    gcode += ";" + n_l
    # ______________________________________________________________________/ header

    # ######################################################################
    # global parameters
    # retraction_on = True  # boolean; is true when retraction is toggled
    fan_on = False  # boolean; is true when fan is toggled
    prev_point = PrintPoint(Point(0, 0, 0), layer_height=1.0,
                            mesh_normal=Vector(1.0, 0.0, 0.0))  # dummy print_point that is overwritten
    layer_height = 0.2  # dummy value that is overwritten
    # ______________________________________________________________________/ global parameters

    # ######################################################################
    # iterate all layers, paths
    print('')
    for point_v, i, j, k in print_organizer.printpoints_indices_iterator():
        layer_height = point_v.layer_height
        # Calculate relative length
        re_l = ((point_v.pt.x - prev_point.pt.x) ** 2 + (point_v.pt.y - prev_point.pt.y) ** 2 + (
                point_v.pt.z - prev_point.pt.z) ** 2) ** 0.5
        if k == 0:  # 'First point
            # retract before moving to first point in path if necessary
            if retraction_min_travel < re_l:
                gcode += "G1 F" + str(feedrate_retraction) + "    ;set retraction feedrate" + n_l
                gcode += "G1" + " E-" + str(retraction_length) + "      ;retract" + n_l
                # ZHOP
                gcode += "G1" + " Z" + '{:.3f}'.format(prev_point.pt.z + z_hop) + "  ;z-hop" + n_l
                # move to first point in path:
                gcode += "G1" + " F" + str(feedrate_travel) + "    ;set travel feedrate" + n_l
                if prev_point.pt.z != point_v.pt.z:
                    gcode += "G1 X" + '{:.3f}'.format(point_v.pt.x) + " Y" + '{:.3f}'.format(point_v.pt.y) + " Z" + '{:.3f}'.format(point_v.pt.z) + n_l
                else:
                    gcode += "G1 X" + '{:.3f}'.format(point_v.pt.x) + " Y" + '{:.3f}'.format(point_v.pt.y) + n_l
                # reverse z-hop after reaching the first point
                gcode += "G1 F" + str(feedrate_retraction) + "    ;set retraction feedrate" + n_l
                gcode += "G1" + " Z" + '{:.3f}'.format(point_v.pt.z) + "  ;reverse z-hop" + n_l
                # reverse retract after reaching the first point
                gcode += "G1" + " E" + str(retraction_length) + "       ;reverse retraction" + n_l
            else:
                if prev_point.pt.z != point_v.pt.z:
                    gcode += "G1 X" + '{:.3f}'.format(point_v.pt.x) + " Y" + '{:.3f}'.format(
                        point_v.pt.y) + " Z" + '{:.3f}'.format(point_v.pt.z) + n_l
                else:
                    gcode += "G1 X" + '{:.3f}'.format(point_v.pt.x) + " Y" + '{:.3f}'.format(point_v.pt.y) + n_l
            # set extrusion feedrate: low for adhesion to bed and normal otherwise
            if point_v.pt.z < min_over_z:
                gcode += "G1" + " F" + str(feedrate_low) + "    ;set low feedrate" + n_l
            else:
                gcode += "G1" + " F" + str(feedrate) + "    ;set extrusion feedrate" + n_l
        else:  # from 2nd point in each path onwards
            # tmpflow = myflow.Branch(b)(i) here we can set the flow multiplier
            # Calculate feedrate : TODO: just a basic formula for now, better ones in the future
            e_val = 4 * re_l * layer_height * path_width / (math.pi * (filament_diameter ** 2))
            if point_v.pt.z < min_over_z:
                e_val *= flow_over
            gcode += "G1 X" + '{:.3f}'.format(point_v.pt.x) + " Y" + '{:.3f}'.format(
                point_v.pt.y) + " E" + '{:.3f}'.format(e_val) + n_l
        prev_point = point_v
        if fan_on is False:
            if i * layer_height >= fan_start_z:  # 'Fan On:
                gcode += "M106 S" + str(fan_speed) + "     ;set fan on to set speed" + n_l
                fan_on = True

    # 'retract after last path
    gcode += "G1 F" + str(feedrate_retraction) + "     ;set ret spd" + n_l
    gcode += "G1" + " E-" + str(retraction_length) + "       ;ret fil" + n_l
    gcode += "G1" + " Z" + '{:.3f}'.format(3 * (prev_point.pt.z + z_hop)) + "  ;ZHop" + n_l
    gcode += "G1 F" + str(feedrate_travel) + "    ;set ret spd" + n_l

    #######################################################################
    # Footer
    gcode += "M201 X500 Y500              ;set acceleration to 500mm/s^2" + n_l
    gcode += "G1" + " F 1000                   ;set feedrate to 1000 mm/min" + n_l
    gcode += "G1 X0 Y0                    ;home x and y axes" + n_l
    gcode += "M104 S0                     ;turn extruder heater off" + n_l
    gcode += "M140 S0                     ;turn bed heater off (if it exists)" + n_l
    gcode += "M84                         ;turn steppers off" + n_l
    gcode += "M106 S0                     ;turn fan off" + n_l
    # ______________________________________________________________________/ footer

    return gcode



__all__ = ['gcode_default_param']


def gcode_default_param(key):
    """ Returns the default parameter with the specified key. """
    if key in default_parameters:
        return default_parameters[key]
    else:
        raise ValueError('The parameter with key : ' + str(key) +
                         ' does not exist in the defaults of gcode parameters. ')


default_parameters = \
    {
        # Physical parameters
        'nozzle_diameter': 0.4,  # in mm
        'filament diameter': 1.75,  # in mm, for calculating E
        'delta': False,  # boolean for delta printers
        'print_volume_x': 300,  # in mm
        'print_volume_y': 300,  # in mm
        'print_volume_z': 600,  # in mm

        # Dimensional parameters
        'layer_width': 0.6,  # in mm

        # Temperature parameters
        'extruder_temperature': 200,  # in °C
        'bed_temperature': 60,  # in °C
        'fan_speed': 255,  # 0-255
        'fan_start_z': 0,  # in mm; height at which the fan starts

        # Movement parameters
        'feedrate': 3600,  # in mm/s
        'feedrate_travel': 4800,  # in mm/s
        'feedrate_low': 1800,  # in mm/s
        'feedrate_retraction': 2400,  # in mm/s
        'acceleration': 0,  # in mm/s²; if set to 0, the default driver value will be used
        'jerk': 0,  # in mm/s; if set to 0, the default driver value will be used

        # Retraction
        'z_hop': 0.5,  # in mm
        'retraction_length': 1,  # in mm
        'retraction_min_travel': 6,  # in mm; below this value, retraction does not happen

        # Adhesion parameters
        'flow_over': 1,  # as fraction, usually > 1; overextrusion value for z < min_over_z, for better adhesion
        'min_over_z': 0,  # in mm; for z < min_over_z, the overextrusion factor applies
    }
